@@ -37,30 +37,27 @@
 
 #include <string.h>
 
-#include <libgupnp/gupnp-control-point.h>
+#include <libgupnp/gupnp.h>
 
+#define SOUP_REQUEST_TIMEOUT 5
 
 struct _GUPnPSimpleIgdPrivate
 {
   GMainContext *main_context;
 
-  GUPnPContext *gupnp_context;
-  GUPnPControlPoint *ip_cp;
-  GUPnPControlPoint *ppp_cp;
+  GUPnPContextManager *gupnp_context_manager;
 
   GPtrArray *service_proxies;
-
   GPtrArray *mappings;
 
-  gulong ip_avail_handler;
-  gulong ip_unavail_handler;
+  gboolean no_new_mappings;
 
-  gulong ppp_avail_handler;
-  gulong ppp_unavail_handler;
+  guint deleting_count;
 };
 
 struct Proxy {
   GUPnPSimpleIgd *parent;
+  GUPnPControlPoint *cp;
   GUPnPServiceProxy *proxy;
 
   gchar *external_ip;
@@ -103,7 +100,6 @@ enum
 enum
 {
   PROP_0,
-  PROP_REQUEST_TIMEOUT,
   PROP_MAIN_CONTEXT
 };
 
@@ -134,9 +130,9 @@ static void gupnp_simple_igd_add_proxy_mapping (GUPnPSimpleIgd *self,
     struct Mapping *mapping);
 
 static void free_proxy (struct Proxy *prox);
-static void free_mapping (struct Mapping *mapping);
+static void free_mapping (GUPnPSimpleIgd *self, struct Mapping *mapping);
 
-static void stop_proxymapping (struct ProxyMapping *pm);
+static void stop_proxymapping (struct ProxyMapping *pm, gboolean stop_renew);
 
 static void gupnp_simple_igd_add_port_real (GUPnPSimpleIgd *self,
     const gchar *protocol,
@@ -172,15 +168,6 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
   klass->add_port = gupnp_simple_igd_add_port_real;
   klass->remove_port = gupnp_simple_igd_remove_port_real;
 
-  g_object_class_install_property (gobject_class,
-      PROP_REQUEST_TIMEOUT,
-      g_param_spec_uint ("request-timeout",
-          "The timeout after which a request is considered to have failed",
-          "After this timeout, the request is considered to have failed and"
-          "is dropped (in seconds).",
-          0, G_MAXUINT, 5,
-          G_PARAM_READWRITE));
-
   g_object_class_install_property (gobject_class,
       PROP_MAIN_CONTEXT,
       g_param_spec_pointer ("main-context",
@@ -247,57 +234,47 @@ gupnp_simple_igd_init (GUPnPSimpleIgd *self)
   self->priv->mappings = g_ptr_array_new ();
 }
 
+/**
+ * gupnp_simple_igd_delete_all_mappings:
+ * @self: a #GUPnPSimpleIgd
+ *
+ * Removes all mappings and prevents other from being formed
+ * Should only be called by the dispose function of subclasses
+ *
+ * Returns: %TRUE if the object can be disposed, %FALSE otherwise
+ */
+
+gboolean
+gupnp_simple_igd_delete_all_mappings (GUPnPSimpleIgd *self)
+{
+  self->priv->no_new_mappings = TRUE;
+
+  while (self->priv->mappings->len)
+  {
+    free_mapping (self, g_ptr_array_index (self->priv->mappings, 0));
+    g_ptr_array_remove_index_fast (self->priv->mappings, 0);
+  }
+
+  return (self->priv->deleting_count == 0);
+}
+
 static void
 gupnp_simple_igd_dispose (GObject *object)
 {
   GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
 
-  if (self->priv->ip_avail_handler)
-    g_signal_handler_disconnect (self->priv->ip_cp,
-        self->priv->ip_avail_handler);
-  self->priv->ip_avail_handler = 0;
+  if (!gupnp_simple_igd_delete_all_mappings (self))
+    return;
 
-  if (self->priv->ip_unavail_handler)
-    g_signal_handler_disconnect (self->priv->ip_cp,
-        self->priv->ip_unavail_handler);
-  self->priv->ip_unavail_handler = 0;
+  if (self->priv->gupnp_context_manager)
+    g_object_unref (self->priv->gupnp_context_manager);
+  self->priv->gupnp_context_manager = NULL;
 
-  if (self->priv->ppp_avail_handler)
-    g_signal_handler_disconnect (self->priv->ppp_cp,
-        self->priv->ppp_avail_handler);
-  self->priv->ppp_avail_handler = 0;
-
-  if (self->priv->ppp_unavail_handler)
-    g_signal_handler_disconnect (self->priv->ppp_cp,
-        self->priv->ppp_unavail_handler);
-  self->priv->ppp_unavail_handler = 0;
-
-  while (self->priv->mappings->len)
-  {
-    free_mapping (
-        g_ptr_array_index (self->priv->mappings, 0));
-    g_ptr_array_remove_index_fast (self->priv->mappings, 0);
+  if (self->priv->service_proxies) {
+    g_ptr_array_foreach (self->priv->service_proxies, (GFunc) free_proxy, NULL);
+    g_ptr_array_free (self->priv->service_proxies, TRUE);
   }
 
-  while (self->priv->service_proxies->len)
-  {
-    free_proxy (
-        g_ptr_array_index (self->priv->service_proxies, 0));
-    g_ptr_array_remove_index_fast (self->priv->service_proxies, 0);
-  }
-
-  if (self->priv->ip_cp)
-    g_object_unref (self->priv->ip_cp);
-  self->priv->ip_cp = NULL;
-
-  if (self->priv->ppp_cp)
-    g_object_unref (self->priv->ppp_cp);
-  self->priv->ppp_cp = NULL;
-
-  if (self->priv->gupnp_context)
-    g_object_unref (self->priv->gupnp_context);
-  self->priv->gupnp_context = NULL;
-
   G_OBJECT_CLASS (gupnp_simple_igd_parent_class)->dispose (object);
 }
 
@@ -335,8 +312,47 @@ _external_ip_address_changed (GUPnPServiceProxy *proxy, const gchar *variable,
 }
 
 static void
-free_proxymapping (struct ProxyMapping *pm)
+_service_proxy_delete_port_mapping (GUPnPServiceProxy *proxy,
+    GUPnPServiceProxyAction *action,
+    gpointer user_data)
 {
+  GError *error = NULL;
+  GUPnPSimpleIgd *self = user_data;
+
+
+  if (!gupnp_service_proxy_end_action (proxy, action, &error,
+          NULL))
+  {
+    g_return_if_fail (error);
+    g_warning ("Error deleting port mapping: %s", error->message);
+  }
+  g_clear_error (&error);
+
+  if (self)
+  {
+    self->priv->deleting_count--;
+    g_object_unref (self);
+  }
+}
+
+static void
+free_proxymapping (struct ProxyMapping *pm, GUPnPSimpleIgd *self)
+{
+  stop_proxymapping (pm, TRUE);
+
+  if (pm->mapped && self)
+  {
+    self->priv->deleting_count++;
+    g_object_ref (self);
+    gupnp_service_proxy_begin_action (pm->proxy->proxy,
+        "DeletePortMapping",
+        _service_proxy_delete_port_mapping, self,
+        "NewRemoteHost", G_TYPE_STRING, "",
+        "NewExternalPort", G_TYPE_UINT, pm->actual_external_port,
+        "NewProtocol", G_TYPE_STRING, pm->mapping->protocol,
+        NULL);
+  }
+
   g_slice_free (struct ProxyMapping, pm);
 }
 
@@ -349,8 +365,6 @@ free_proxy (struct Proxy *prox)
   gupnp_service_proxy_remove_notify (prox->proxy, "ExternalIPAddress",
       _external_ip_address_changed, prox);
 
-  g_object_unref (prox->proxy);
-  g_ptr_array_foreach (prox->proxymappings, (GFunc) stop_proxymapping, NULL);
   g_ptr_array_foreach (prox->proxymappings, (GFunc) free_proxymapping, NULL);
   g_ptr_array_free (prox->proxymappings, TRUE);
   g_free (prox->external_ip);
@@ -358,8 +372,26 @@ free_proxy (struct Proxy *prox)
 }
 
 static void
-free_mapping (struct Mapping *mapping)
+free_mapping (GUPnPSimpleIgd *self, struct Mapping *mapping)
 {
+  guint i, j;
+
+  for (i=0; i < self->priv->service_proxies->len; i++)
+  {
+    struct Proxy *prox = g_ptr_array_index (self->priv->service_proxies, i);
+
+    for (j=0; j < prox->proxymappings->len; j++)
+    {
+      struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, j);
+      if (pm->mapping == mapping)
+      {
+        free_proxymapping (pm, self);
+        g_ptr_array_remove_index_fast (prox->proxymappings, j);
+        j--;
+      }
+    }
+  }
+
   g_free (mapping->protocol);
   g_free (mapping->local_ip);
   g_free (mapping->description);
@@ -373,9 +405,6 @@ gupnp_simple_igd_finalize (GObject *object)
 
   g_main_context_unref (self->priv->main_context);
 
-  g_warn_if_fail (self->priv->service_proxies->len == 0);
-  g_ptr_array_free (self->priv->service_proxies, TRUE);
-
   g_warn_if_fail (self->priv->mappings->len == 0);
   g_ptr_array_free (self->priv->mappings, TRUE);
 
@@ -389,13 +418,6 @@ gupnp_simple_igd_get_property (GObject *object, guint prop_id,
   GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
 
   switch (prop_id) {
-    case PROP_REQUEST_TIMEOUT:
-      {
-        SoupSession *session;
-        session = gupnp_context_get_session (self->priv->gupnp_context);
-        g_object_get_property (G_OBJECT (session), "timeout", value);
-      }
-      break;
     case PROP_MAIN_CONTEXT:
       g_value_set_pointer (value, self->priv->main_context);
       break;
@@ -413,13 +435,6 @@ gupnp_simple_igd_set_property (GObject *object, guint prop_id,
   GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
 
   switch (prop_id) {
-    case PROP_REQUEST_TIMEOUT:
-      {
-        SoupSession *session;
-        session = gupnp_context_get_session (self->priv->gupnp_context);
-        g_object_set_property (G_OBJECT (session), "timeout", value);
-      }
-      break;
     case PROP_MAIN_CONTEXT:
       if (!self->priv->main_context && g_value_get_pointer (value))
       {
@@ -438,11 +453,17 @@ _cp_service_avail (GUPnPControlPoint *cp,
     GUPnPServiceProxy *proxy,
     GUPnPSimpleIgd *self)
 {
-  struct Proxy *prox = g_slice_new0 (struct Proxy);
+  struct Proxy *prox;
   guint i;
 
+  if (self->priv->no_new_mappings)
+    return;
+
+  prox = g_slice_new0 (struct Proxy);
+
   prox->parent = self;
-  prox->proxy = g_object_ref (proxy);
+  prox->cp = cp;
+  prox->proxy = proxy;
   prox->proxymappings = g_ptr_array_new ();
 
   gupnp_simple_igd_gather (self, prox);
@@ -464,16 +485,12 @@ _cp_service_unavail (GUPnPControlPoint *cp,
 
   for (i=0; i < self->priv->service_proxies->len; i++)
   {
-    struct Proxy *prox =
-      g_ptr_array_index (self->priv->service_proxies, i);
+    struct Proxy *prox = g_ptr_array_index (self->priv->service_proxies, i);
 
-    if (!strcmp (gupnp_service_info_get_udn (GUPNP_SERVICE_INFO (prox->proxy)),
+    if (prox->cp == cp &&
+        !strcmp (gupnp_service_info_get_udn (GUPNP_SERVICE_INFO (proxy)),
             gupnp_service_info_get_udn (GUPNP_SERVICE_INFO (prox->proxy))))
     {
-      g_ptr_array_foreach (prox->proxymappings, (GFunc) stop_proxymapping,
-          NULL);
-      g_ptr_array_foreach (prox->proxymappings, (GFunc) free_proxymapping,
-          NULL);
       free_proxy (prox);
       g_ptr_array_remove_index_fast (self->priv->service_proxies, i);
       break;
@@ -481,6 +498,41 @@ _cp_service_unavail (GUPnPControlPoint *cp,
   }
 }
 
+static void
+gupnp_simple_igd_add_control_point (GUPnPSimpleIgd *self,
+    GUPnPContext *gupnp_context, const char *target)
+{
+  GUPnPControlPoint *cp;
+
+  cp = gupnp_control_point_new (gupnp_context, target);
+  g_return_if_fail (cp);
+
+  g_signal_connect_object (cp, "service-proxy-available",
+      G_CALLBACK (_cp_service_avail), self, 0);
+  g_signal_connect_object (cp, "service-proxy-unavailable",
+      G_CALLBACK (_cp_service_unavail), self, 0);
+
+  gssdp_resource_browser_set_active (GSSDP_RESOURCE_BROWSER (cp), TRUE);
+
+  gupnp_context_manager_manage_control_point (
+      self->priv->gupnp_context_manager, cp);
+}
+
+static void
+_context_available (GUPnPContextManager *manager, GUPnPContext *gupnp_context,
+    GUPnPSimpleIgd *self)
+{
+  SoupSession *session;
+
+  session = gupnp_context_get_session (gupnp_context);
+  g_object_set (session, "timeout", SOUP_REQUEST_TIMEOUT, NULL);
+
+  gupnp_simple_igd_add_control_point (self, gupnp_context,
+      "urn:schemas-upnp-org:service:WANIPConnection:1");
+  gupnp_simple_igd_add_control_point (self, gupnp_context,
+      "urn:schemas-upnp-org:service:WANPPPConnection:1");
+}
+
 
 static void
 gupnp_simple_igd_constructed (GObject *object)
@@ -491,42 +543,11 @@ gupnp_simple_igd_constructed (GObject *object)
   if (!self->priv->main_context)
     self->priv->main_context = g_main_context_ref (g_main_context_default ());
 
-  self->priv->gupnp_context = gupnp_context_new (self->priv->main_context,
-      NULL, 0, NULL);
-  g_return_if_fail (self->priv->gupnp_context);
+  self->priv->gupnp_context_manager =
+    gupnp_context_manager_new (self->priv->main_context, 0);
 
-  session = gupnp_context_get_session (self->priv->gupnp_context);
-  g_object_set (session, "timeout", 5, NULL);
-
-  self->priv->ip_cp = gupnp_control_point_new (self->priv->gupnp_context,
-      "urn:schemas-upnp-org:service:WANIPConnection:1");
-  g_return_if_fail (self->priv->ip_cp);
-
-  self->priv->ip_avail_handler = g_signal_connect (self->priv->ip_cp,
-      "service-proxy-available",
-      G_CALLBACK (_cp_service_avail), self);
-  self->priv->ip_unavail_handler = g_signal_connect (self->priv->ip_cp,
-      "service-proxy-unavailable",
-      G_CALLBACK (_cp_service_unavail), self);
-
-  self->priv->ppp_cp = gupnp_control_point_new (self->priv->gupnp_context,
-      "urn:schemas-upnp-org:service:WANPPPConnection:1");
-  g_return_if_fail (self->priv->ppp_cp);
-
-  self->priv->ppp_avail_handler = g_signal_connect (self->priv->ppp_cp,
-      "service-proxy-available",
-      G_CALLBACK (_cp_service_avail), self);
-  self->priv->ppp_unavail_handler = g_signal_connect (self->priv->ppp_cp,
-      "service-proxy-unavailable",
-      G_CALLBACK (_cp_service_unavail), self);
-
-
-  gssdp_resource_browser_set_active (
-      GSSDP_RESOURCE_BROWSER (self->priv->ip_cp),
-      TRUE);
-  gssdp_resource_browser_set_active (
-      GSSDP_RESOURCE_BROWSER (self->priv->ppp_cp),
-      TRUE);
+  g_signal_connect (self->priv->gupnp_context_manager, "context-available",
+      G_CALLBACK (_context_available), self);
 
   if (G_OBJECT_CLASS (gupnp_simple_igd_parent_class)->constructed)
     G_OBJECT_CLASS (gupnp_simple_igd_parent_class)->constructed (object);
@@ -653,7 +674,10 @@ static void
 gupnp_simple_igd_call_add_port_mapping (struct ProxyMapping *pm,
     GUPnPServiceProxyActionCallback callback)
 {
+  g_assert (pm);
   g_return_if_fail (pm->action == NULL);
+  g_assert (pm->proxy);
+  g_assert (pm->mapping);
 
   pm->action = gupnp_service_proxy_begin_action (pm->proxy->proxy,
       "AddPortMapping",
@@ -674,7 +698,7 @@ _renew_mapping_timeout (gpointer user_data)
 {
   struct ProxyMapping *pm = user_data;
 
-  stop_proxymapping (pm);
+  stop_proxymapping (pm, FALSE);
 
   gupnp_simple_igd_call_add_port_mapping (pm,
       _service_proxy_renewed_port_mapping);
@@ -708,12 +732,14 @@ _service_proxy_added_port_mapping (GUPnPServiceProxy *proxy,
 
 
 
-    pm->renew_src =
-      g_timeout_source_new_seconds (pm->mapping->lease_duration / 2);
-    g_source_set_callback (pm->renew_src,
-        _renew_mapping_timeout, pm, NULL);
-    g_source_attach (pm->renew_src, self->priv->main_context);
-
+    if (pm->mapping->lease_duration > 0)
+    {
+      pm->renew_src =
+        g_timeout_source_new_seconds (pm->mapping->lease_duration / 2);
+      g_source_set_callback (pm->renew_src,
+          _renew_mapping_timeout, pm, NULL);
+      g_source_attach (pm->renew_src, self->priv->main_context);
+    }
   }
   else
   {
@@ -848,31 +874,13 @@ gupnp_simple_igd_add_port (GUPnPSimpleIgd *self,
       lease_duration, description);
 }
 
-
-static void
-_service_proxy_delete_port_mapping (GUPnPServiceProxy *proxy,
-    GUPnPServiceProxyAction *action,
-    gpointer user_data)
-{
-  GError *error = NULL;
-
-
-  if (!gupnp_service_proxy_end_action (proxy, action, &error,
-          NULL))
-  {
-    g_return_if_fail (error);
-    g_warning ("Error deleting port mapping: %s", error->message);
-  }
-  g_clear_error (&error);
-}
-
 static void
 gupnp_simple_igd_remove_port_real (GUPnPSimpleIgd *self,
     const gchar *protocol,
     guint external_port)
 {
-  guint i, j;
   struct Mapping *mapping = NULL;
+  guint i;
 
   g_return_if_fail (protocol);
 
@@ -891,41 +899,7 @@ gupnp_simple_igd_remove_port_real (GUPnPSimpleIgd *self,
 
   g_ptr_array_remove_index_fast (self->priv->mappings, i);
 
-  for (i=0; i < self->priv->service_proxies->len; i++)
-  {
-    struct Proxy *prox = g_ptr_array_index (self->priv->service_proxies, i);
-
-    for (j=0; j < prox->proxymappings->len; j++)
-    {
-      struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, j);
-      if (pm->mapping == mapping)
-      {
-        stop_proxymapping (pm);
-
-        if (pm->renew_src)
-        {
-          g_source_destroy (pm->renew_src);
-          g_source_unref (pm->renew_src);
-        }
-        pm->renew_src = NULL;
-
-        if (pm->mapped)
-          gupnp_service_proxy_begin_action (prox->proxy,
-              "DeletePortMapping",
-              _service_proxy_delete_port_mapping, self,
-              "NewRemoteHost", G_TYPE_STRING, "",
-              "NewExternalPort", G_TYPE_UINT, pm->actual_external_port,
-              "NewProtocol", G_TYPE_STRING, mapping->protocol,
-              NULL);
-
-        free_proxymapping (pm);
-        g_ptr_array_remove_index_fast (prox->proxymappings, j);
-        j--;
-      }
-    }
-  }
-
-  free_mapping (mapping);
+  free_mapping (self, mapping);
 }
 
 /**
@@ -954,10 +928,17 @@ gupnp_simple_igd_remove_port (GUPnPSimpleIgd *self,
 }
 
 static void
-stop_proxymapping (struct ProxyMapping *pm)
+stop_proxymapping (struct ProxyMapping *pm, gboolean stop_renew)
 {
   if (pm->action)
     gupnp_service_proxy_cancel_action (pm->proxy->proxy,
         pm->action);
   pm->action = NULL;
+
+  if (stop_renew && pm->renew_src)
+  {
+    g_source_destroy (pm->renew_src);
+    g_source_unref (pm->renew_src);
+    pm->renew_src = NULL;
+  }
 }
